float Script : STANDARDSGLOBAL <
    string   UiControl = "none";
    string   ScriptClass = "scene";
    string   ScriptOrder = "standard";
    string   ScriptOutput = "colour";
    string   Script = "Technique=PBR_Metal_Rough_Cube;";
    string   ScriptDescription = "<b>PBR that uses metalness and roughness maps with cube environment maps</b>";
> = 0.1; // version #

int nbSamples
<
    string   UiName = "nbSamples";
    string   UiControl = "Slider";
   float UiMin = 4.0f;
   float UiMax = 64.0f;
 > = 16;
 
float envExposure
<
    string   UiName = "Env Exposure";
    string   UiControl = "Slider";
   float UiMin = -10.0f;
   float UiMax = 10.0f;
 > = 1;
 
float envRotation
<
    string   UiName = "Env Rotation";
    string   UiControl = "Slider";
   float UiMin = 0.0f;
   float UiMax = 360.0f;
 > = 0;

sampler2D tex0 <  string BindAddress = "Tog.Texture[0].2D"; >;			///albedo
samplerCUBE tex1 <  string BindAddress = "Tog.Texture[1].cube"; >;		///Environment
sampler2D tex2 <  string BindAddress = "Tog.Texture[2].2D"; >;			///metal/rough/AO (rgb)
sampler2D tex3 <  string BindAddress = "Tog.Texture[3].2D"; >;			///Normal
float4x4 worldMatrix <	 string BindAddress = "Tog.Camera.Model"; >;
float4x4 invTransWorldMatrix <  string BindAddress = "Tog.Camera.InverseTransposeModel";>;
float4x4 inverseViewMatrix <  string BindAddress = "Tog.Camera.InverseView";>;

const float M_PI = 3.1415926535897932384626433832795;
const float M_INV_PI = 0.31830988618379067153776752674503;
const float M_INV_LOG2 = 1.4426950408889634073599246810019;
const int maxLod = 11;

float radicalInverse[64] = {0, 0.5, 0.25, 0.75, 0.125, 0.625, 0.375, 0.875, 0.0625, 0.5625, 0.3125, 0.8125, 0.1875, 0.6875, 0.4375, 0.9375, 0.03125, 0.53125, 0.28125, 0.78125, 0.15625, 0.65625, 0.40625, 0.90625, 0.09375, 0.59375, 0.34375, 0.84375, 0.21875, 0.71875, 0.46875, 0.96875, 0.015625, 0.515625, 0.265625, 0.765625, 0.140625, 0.640625, 0.390625, 0.890625, 0.078125, 0.578125, 0.328125, 0.828125, 0.203125, 0.703125, 0.453125, 0.953125, 0.046875, 0.546875, 0.296875, 0.796875, 0.171875, 0.671875, 0.421875, 0.921875, 0.109375, 0.609375, 0.359375, 0.859375, 0.234375, 0.734375, 0.484375, 0.984375};

struct State
{
	float4 tex_coord;
	float3 positionOS;
	float3 origin;
	float3 normalOS;
	float3 tangentOS;
	float3 bitangentOS;
	float3 direction;
	float ray_length;
	float3x3 tangent_space;
};

struct App2vert
{
	float3 position : POSITION;
	float3 normal : NORMAL;
	float3 tangent : TANGENT;
	float3 bitangent : BINORMAL;	
	float4 texcoord0 : TEXCOORD0;
	float4 texcoord1 : TEXCOORD1;
	float4 texcoord2 : TEXCOORD2;
	float4 texcoord3 : TEXCOORD3;
};

struct Vert2frag
{
	float4 hpos : POSITION;
	float4 tex_coord : TEXCOORD0;
	float4 position : TEXCOORD6;
	float3 normal : TEXCOORD7;
	float3 tangent : TEXCOORD4;
	float3 bitangent : TEXCOORD5;	
};

Vert2frag vertex_main(
	App2vert vs_in)
{
	Vert2frag vs_out;
	float4 position = float4(vs_in.position, 1);
	//WVP * position
	vs_out.hpos = mul(glstate.matrix.mvp, position);
	//position
	vs_out.position = position;
	//normal, tangent, bitangent
	vs_out.normal = normalize(vs_in.normal);
	vs_out.tangent = normalize(vs_in.tangent);
	vs_out.bitangent = normalize(vs_in.bitangent);
	//texcoord
	vs_out.tex_coord = vs_in.texcoord0;
	return vs_out;
}

typedef float4 FS_OUT;

void PBR_Inputs(
	State in_state,
	sampler2D in_albedoTexture,
	float in_albedoMultiplier,
	sampler2D in_normalTexture,
	float in_normalMultiplier,
	sampler2D in_aoTexture,
	float in_aoMultiplier,
	sampler2D in_roughnessTexture,
	float in_roughnessMultiplier,
	sampler2D in_metalnessTexture,
	float in_metalnessMultiplier,
	out float4 out_albedo,
	out float3 out_normal,
	out float out_ao,
	out float out_roughness,
	out float out_metalness
	)	
{
	//float4 of (2.2, 2.2, 2.2, 1.0) used to transform sampled texels from Gamma>Linear space
	float4 gammaFactor = float4(2.2, 2.2, 2.2, 1.0);//2.2
		
	//gets Gamma space Albedo colour for this texel, Gamma>Linear space transform, scale by Albedo Multiplier
	float4 albedo = tex2D(in_albedoTexture, in_state.tex_coord.xy);
	albedo = pow(albedo, gammaFactor);
	out_albedo = albedo * in_albedoMultiplier;
				
	//gets normal map value for this texel, transforms it to tangent space
	float3 normal = tex2D(in_normalTexture, in_state.tex_coord.xy).xyz;
	normal = (normal.xyz - 0.5) * 2.0;
	normal.xyz *= in_normalMultiplier;
	out_normal = normal;
		
	//gets gamma space Ambient Occlusion value of this texel and gamma>linear
	float ao = tex2D(in_aoTexture, in_state.tex_coord.xy).r;
	out_ao = ao * in_aoMultiplier;
		
	//gets gamma space roughness value for this texel and gamma>linear
	float roughness = tex2D(in_roughnessTexture, in_state.tex_coord.xy).b;
	out_roughness = roughness * in_roughnessMultiplier;
		
	//gets gamma metalness
	//metalness gamma>linear
	//scales by in_metalnessMultiplier
	float metalness = tex2D(in_metalnessTexture, in_state.tex_coord.xy).g;
	out_metalness = metalness * in_metalnessMultiplier;
}

float normal_distrib(
	float ndh,
	float Roughness){
  // use GGX / Trowbridge-Reitz, same as Disney and Unreal 4
  // cf http://blog.selfshadow.com/publications/s2013-shading-course/karis/s2013_pbs_epic_notes_v2.pdf p3
  float alpha = Roughness * Roughness;
  float tmp = alpha / max(1e-8,(ndh*ndh*(alpha*alpha-1.0)+1.0));
  return tmp * tmp * M_INV_PI;
}

float3 normalBlend(
	float3 normA,
	float3 normB){
	return normalize(float3(normA.x + normB.x, normA.y + normB.y, normA.z * normB.z));
}

float probabilityGGX(
	float ndh,
	float vdh,
	float Roughness){
  return normal_distrib(ndh, Roughness) * ndh / (4.0*vdh);
}

float distortion(
	float3 Wn){
  // Computes the inverse of the solid angle of the (differential) pixel in
  // the cube map pointed at by Wn
  float sinT = sqrt(1.0-Wn.y*Wn.y);
  return sinT;
}

float computeLOD(
	float3 Ln,
	float p){
  return max(0.0, (maxLod-1.5) - 0.5*(log(float(nbSamples)) + log( p * distortion(Ln) ))	* M_INV_LOG2);
}

float3 envSampleLOD(
	float3 dirWS,
	float lod){
	
	//spherical mapping
	/*float n = length(dirWS.xz);
	float2 pos = float2( (n>0.0000001) ? dirWS.x / n : 0.0, dirWS.y);
	pos = acos(pos)*M_INV_PI;
	pos.x = (dirWS.z > 0.0) ? pos.x*0.5 : 1.0-(pos.x*0.5);
	pos.x += envRotation;
	pos.y = 1.0-pos.y;
	return tex2D(tex1, pos).rgb * envExposure;*/

	//cube mapping
	float3 rotatedUVW = dirWS;
	rotatedUVW.x = (cos(radians(envRotation)) * dirWS.x) + (sin(radians(envRotation)) * dirWS.z);
	rotatedUVW.y = dirWS.y;
	rotatedUVW.z = (-1 * (sin(radians(envRotation)) * dirWS.x)) + (cos(radians(envRotation)) * dirWS.z);
	return texCUBElod(tex1, float4(rotatedUVW.xyz, max(1, lod))).rgb * envExposure;
}

float G1(
	float ndw, // w is either Ln or Vn
	float k){
  // One generic factor of the geometry function divided by ndw
  // NB : We should have k > 0
  return 1.0 / ( ndw*(1.0-k) +  k );
}

float3 fresnel(
	float vdh,
	float3 F0)
{
  // Schlick with Spherical Gaussian approximation
  // cf http://blog.selfshadow.com/publications/s2013-shading-course/karis/s2013_pbs_epic_notes_v2.pdf p3
  float sphg = pow(2.0, (-5.55473*vdh - 6.98316) * vdh);
  return F0 + (vec3(1.0, 1.0, 1.0) - F0) * sphg;
}

float visibility(
	float ndl,
	float ndv,
	float Roughness){
  // Schlick with Smith-like choice of k
  // cf http://blog.selfshadow.com/publications/s2013-shading-course/karis/s2013_pbs_epic_notes_v2.pdf p3
  // visibility is a Cook-Torrance geometry function divided by (n.l)*(n.v)
  float k = max(Roughness * Roughness * 0.5, 1e-5);
  return G1(ndl,k)*G1(ndv,k);
}

float3 cook_torrance_contrib(
	float vdh,
	float ndh,
	float ndl,
	float ndv,
	float3 Ks,
	float Roughness){
  // This is the contribution when using importance sampling with the GGX based
  // sample distribution. This means ct_contrib = ct_brdf / ggx_probability
  return fresnel(vdh,Ks) * (visibility(ndl,ndv,Roughness) * vdh * ndl / ndh );
}

float3 importanceSampleGGX(
	float2 Xi,
	float3 A,
	float3 B,
	float3 C,
	float roughness){
	float a = roughness*roughness;
	float cosT = sqrt((1.0-Xi.y)/(1.0+(a*a-1.0)*Xi.y));
	float sinT = sqrt(1.0-cosT*cosT);
	float phi = 2.0*M_PI*Xi.x;
	return (sinT*cos(phi)) * A + (sinT*sin(phi)) * B + cosT * C;
}

float radicalInverse_VdC(
uint bits){
     bits = (bits << 16u) | (bits >> 16u);
     bits = ((bits & 0x55555555u) << 1u) | ((bits & 0xAAAAAAAAu) >> 1u);
     bits = ((bits & 0x33333333u) << 2u) | ((bits & 0xCCCCCCCCu) >> 2u);
     bits = ((bits & 0x0F0F0F0Fu) << 4u) | ((bits & 0xF0F0F0F0u) >> 4u);
     bits = ((bits & 0x00FF00FFu) << 8u) | ((bits & 0xFF00FF00u) >> 8u);
     return float(bits) * 2.3283064365386963e-10; // / 0x100000000
}

float2 hammersley2D(
	int i,
	int nbSamples){
	return float2((float)i / (float)nbSamples, radicalInverse[i]);
}

float3 PBR_Main(
	State in_state,
	sampler2D in_baseColour,
	samplerCUBE in_environmentMap,
	sampler2D in_normalMap,
	sampler2D in_AO_Metal_Rough){
	
	float ao, roughness, metallic;
	float3 normalMap, baseColour;
	float4 out_colour;
	
	PBR_Inputs(in_state, in_baseColour, 1, in_normalMap, 1, in_AO_Metal_Rough, 1, in_AO_Metal_Rough, 1, in_AO_Metal_Rough, 1, out_colour, normalMap, ao, roughness, metallic);
		baseColour = out_colour.rgb;
	
	float3 eyeVecVS = float3(0.0,0.0,0.0) - normalize(mul((float3x3)glstate.matrix.modelview[0], in_state.positionOS));
	float3 eyeVecWS = normalize(mul(inverseViewMatrix, float4(0,0,0,1)) - mul(worldMatrix, float4(in_state.positionOS.xyz, 1))).xyz;
	
	float3 normalTS = normalMap;
	
	float3 normalMapOS = normalize(
	normalTS.x * in_state.tangentOS
	 + normalTS.y * in_state.bitangentOS
	 + normalTS.z * in_state.normalOS
	);
	
	float3 normalMapVS = normalize(mul((float3x3)glstate.matrix.invtrans.modelview[0], normalMapOS));
	float3 normalMapWS = normalize(mul((float3x3)invTransWorldMatrix, normalMapOS));
		
	float3 dielectricColour = float3(0.04);
	
	ao = lerp(1.0, ao, 1.0 * (1.0 - metallic));

	//balances the base colour between diff and spec based on metalness
	float3 diffColour = baseColour * (1.0 - metallic);
	float3 specColour = lerp(dielectricColour, baseColour, metallic);
	  
    float3 TpOS = normalize(
	in_state.tangentOS
	- (normalMapOS * dot(in_state.tangentOS, normalMapOS))
  ); // local tangent
  float3 BpOS = normalize(
	in_state.bitangentOS
	- (normalMapOS * dot(in_state.bitangentOS, normalMapOS))
	- (TpOS * dot(in_state.bitangentOS, TpOS))
  ); // local bitangent

  float ndv = dot(eyeVecWS, normalMapWS);

    if (ndv < 0) {
	eyeVecWS = reflect(eyeVecWS, normalMapWS);
	ndv = abs(ndv);
  }

    float3 contribE = /*envIrradiance(normalWS) */ diffColour * ao;//important, probably the enviorment mapping diffuse or something

	float3 contribS = float3(0.0);
  for(int i=0; i<nbSamples; ++i){
	float2 Xi = hammersley2D(i, nbSamples);
	float3 HnOS = importanceSampleGGX(Xi,TpOS,BpOS,normalMapOS,roughness);
	float3 HnWS = mul((float3x3)worldMatrix, HnOS);
	float3 LnWS = -reflect(normalize(eyeVecWS), normalize(HnWS));
	
	float ndl = dot(normalMapWS, LnWS);
	ndl = max( 1e-8, abs(ndl) );

	// Horizon fading trick from http://marmosetco.tumblr.com/post/81245981087
	const float horizonFade = 1.3;
	float horiz = clamp( 1.0 + horizonFade * ndl, 0.0, 1.0 );
	horiz *= horiz;
	
	float vdh = max(1e-8, dot(normalize(eyeVecWS), normalize(HnWS)));
	float ndh = max(1e-8, dot(normalize(normalMapWS), normalize(HnWS)));
	float prob = probabilityGGX(ndh, vdh, roughness);
	float lodS = roughness < 0.01 ? 0.0 : computeLOD(LnWS, probabilityGGX(ndh, vdh, roughness));
	
	contribS += envSampleLOD(LnWS, lodS) * cook_torrance_contrib(vdh, ndh, ndl, ndv, specColour, roughness) * horiz * ao;
  }
  contribS /= float(nbSamples);

  // Emissive
  float3 contribEm =  0/*emissive_intensity * texture2D(emissive_tex, in_state.tex_coord).rgb*/;//not yet used

  // Sum diffuse + spec + emissive
  return contribS + contribE + contribEm;
}


FS_OUT fragment_main(
	Vert2frag fs_in) : COLOR
{
	State state;
	
	state.positionOS = fs_in.position.xyz;	
	state.normalOS = fs_in.normal;
	state.tangentOS = fs_in.tangent;
	state.bitangentOS = fs_in.bitangent;	
	state.tex_coord = fs_in.tex_coord;
	state.origin = 0;
	state.ray_length = length(state.positionOS);
	state.direction = state.positionOS/state.ray_length;

	//currently not needed
	/*state.tangent_space[0][0] = state.tangentOS.x;
	state.tangent_space[1][0] = state.tangentOS.y;
	state.tangent_space[2][0] = state.tangentOS.z;
	state.tangent_space[0][1] = state.bitangentOS.x;
	state.tangent_space[1][1] = state.bitangentOS.y;
	state.tangent_space[2][1] = state.bitangentOS.z;
	state.tangent_space[0][2] = state.normalOS.x;
	state.tangent_space[1][2] = state.normalOS.y;
	state.tangent_space[2][2] = state.normalOS.z;*/
	
	FS_OUT fs_out;
	fs_out = float4(pow(PBR_Main(state, tex0, tex1, tex3, tex2), (1 / 2.2)), 1);
	return fs_out;
}

technique PBR_Metal_Rough_Cube
{
	pass P0
	{
		DepthTestEnable = true;
		DepthMask       = true;
		CullFaceEnable  = true;
		BlendEnable = true;
		blendfunc = int2(srcalpha, oneminussrcalpha);
		VertexShader = compile vp40 vertex_main();
		PixelShader  = compile gp4fp fragment_main(); // gp4fp
	}
}

 string description = "PBRTestConversion";